package gravity;

import java.nio.ByteBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.LWJGLException;
import org.lwjgl.input.Keyboard;
import org.lwjgl.opengl.ContextAttribs;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL15;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;
import org.lwjgl.opengl.PixelFormat;
import org.lwjgl.util.vector.Matrix4f;
import org.lwjgl.util.vector.Vector3f;
import org.lwjgl.util.vector.Vector4f;

public class GravityDisplay {
	
	static {
		String os = System.getProperty("os.name").toLowerCase();
		String path;
		if (os.contains("win")) path = "windows";
		else if (os.contains("mac")) path = "macosx";
		else if (os.contains("sol")) path = "solaris";
		else path = "linux";
		String libPath = GravityDisplay.class.getResource("/native/" + path).getPath();
		if (libPath != null) System.setProperty("org.lwjgl.librarypath", libPath);
	}
	
	public static int width = 960, height = 540;
	public static final int bufferSize = 100 * 7;
	private static final Vector3f unitX = new Vector3f(1, 0, 0), unitY = new Vector3f(0, 1, 0), unitZ = new Vector3f(0, 0, 1);
	
	private final Simulation simulation;
	
	private Matrix4f projMat, viewMat;
	private Vector3f position;
	
	private int vaoID = 0;
	private int vboID = 0;
	private ByteBuffer buffer;
	
	private Thread displayThread;
	protected Object displayUpdateObject = new Object();
	
	protected GravityDisplay(Simulation simulation) {
		this.simulation = simulation;
		
		displayThread = new Thread(() -> {
			initGL();
			init();
			loop();
		});
		displayThread.start();
	}
	
	private void init() {
		projMat = projectionMatrix(70, 16f / 9f, 0.1f, 1000f, null);
		viewMat = new Matrix4f();
		position = new Vector3f(0, 0, 2);
		Matrix4f.translate(position.negate(null), viewMat, viewMat);
		Shader.init();
		Shader.particle.enable();
		Shader.particle.setUniformMat4("projMat", projMat);
		
		vaoID = GL30.glGenVertexArrays();
		vboID = GL15.glGenBuffers();
		
		GL30.glBindVertexArray(vaoID);
		buffer = BufferUtils.createByteBuffer((3 * 4 + 3 * 4 + 4) * bufferSize);
		GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
		GL15.glBufferData(GL15.GL_ARRAY_BUFFER, buffer.limit(), GL15.GL_STREAM_DRAW);
		
		GL20.glEnableVertexAttribArray(0);
		GL20.glVertexAttribPointer(0, 3, GL11.GL_FLOAT, false, 6 * 4 + 4, 0);// position
		GL20.glEnableVertexAttribArray(1);
		GL20.glVertexAttribPointer(1, 3, GL11.GL_FLOAT, false, 6 * 4 + 4, 3 * 4);// color
		GL20.glEnableVertexAttribArray(2);
		GL20.glVertexAttribPointer(2, 1, GL11.GL_FLOAT, false, 6 * 4 + 4, 6 * 4);// mass
	}
	
	private void render() {
		final double[] pos = simulation.position, vel = simulation.velocity, mass = simulation.mass;
		
		Shader.particle.enable();
		for (int j = 0, len = simulation.numParticles / bufferSize + (simulation.numParticles % bufferSize == 0 ? 0 : 1); j < len; j++) {
			buffer.clear();
			int iLen = j < len - 1 || simulation.numParticles % bufferSize == 0 ? bufferSize : simulation.numParticles % bufferSize;
			for (int i = 0; i < iLen; i++) {
				buffer.putFloat((float) pos[(j * bufferSize + i) * 3 + 0]);
				buffer.putFloat((float) pos[(j * bufferSize + i) * 3 + 1]);
				buffer.putFloat((float) pos[(j * bufferSize + i) * 3 + 2]);
				buffer.putFloat((float) Math.abs(vel[(j * bufferSize + i) * 3 + 0]) * 10 + 0.5f);
				buffer.putFloat((float) Math.abs(vel[(j * bufferSize + i) * 3 + 1]) * 10 + 0.5f);
				buffer.putFloat((float) Math.abs(vel[(j * bufferSize + i) * 3 + 2]) * 10 + 0.5f);
				buffer.putFloat((float) mass[j * bufferSize + i]);
			}
			buffer.flip();
			GL15.glBindBuffer(GL15.GL_ARRAY_BUFFER, vboID);
			GL15.glBufferData(GL15.GL_ARRAY_BUFFER, buffer.limit(), GL15.GL_STREAM_DRAW);
			GL15.glBufferSubData(GL15.GL_ARRAY_BUFFER, 0, buffer);
			
			GL30.glBindVertexArray(vaoID);
			GL11.glDrawArrays(GL11.GL_POINTS, 0, iLen);
		}
	}
	
	private void loop() {
		Vector3f move = new Vector3f();
		Vector4f rotationAdd = new Vector4f(), move4f = new Vector4f();
		Matrix4f rotAddMat = new Matrix4f(), viewMatInv = new Matrix4f();
		
		int delta, ticks = 0, fps = 0;
		boolean freeRotation = false, R_down = false, I_down = false;
		long prevTime = System.nanoTime(), prevSecond = prevTime, curTime;
		
		while (!Display.isCloseRequested() && simulation.running) {
			delta = (int) (((curTime = System.nanoTime()) - prevTime) * 1e-9 * 60);
			
			for (int d = 0; d < delta; d++) {
				if (simulation.playingSimulation) simulation.loadDataFromInput();
				
				float speed = Keyboard.isKeyDown(Keyboard.KEY_LCONTROL) ? 0.002f : Keyboard.isKeyDown(Keyboard.KEY_RCONTROL) ? 1f : 0.02f;
				
				rotationAdd.set(0, 0, 0, 1);
				if (Keyboard.isKeyDown(Keyboard.KEY_LEFT)) rotationAdd.y += speed;
				if (Keyboard.isKeyDown(Keyboard.KEY_RIGHT)) rotationAdd.y += -speed;
				if (Keyboard.isKeyDown(Keyboard.KEY_UP)) rotationAdd.x += speed;
				if (Keyboard.isKeyDown(Keyboard.KEY_DOWN)) rotationAdd.x += -speed;
				
				move.set(0, 0, 0);
				if (Keyboard.isKeyDown(Keyboard.KEY_W)) move.z--;
				if (Keyboard.isKeyDown(Keyboard.KEY_S)) move.z++;
				if (Keyboard.isKeyDown(Keyboard.KEY_A)) move.x--;
				if (Keyboard.isKeyDown(Keyboard.KEY_D)) move.x++;
				if (Keyboard.isKeyDown(Keyboard.KEY_SPACE)) move.y++;
				if (Keyboard.isKeyDown(Keyboard.KEY_LSHIFT)) move.y--;
				move.scale(speed);
				if (Keyboard.isKeyDown(Keyboard.KEY_R)) {
					if (!R_down) {
						R_down = true;
						freeRotation = !freeRotation;
					}
				} else R_down = false;
				if (Keyboard.isKeyDown(Keyboard.KEY_F)) {
					try {
						boolean fullscreen = Display.isFullscreen();
						Display.destroy();
						if (!fullscreen) {
							Display.setDisplayModeAndFullscreen(Display.getDesktopDisplayMode());
							Display.setFullscreen(true);
						} else {
							Display.setDisplayMode(new DisplayMode(width, height));
							Display.setFullscreen(false);
						}
						Display.create(new PixelFormat().withAlphaBits(8).withSamples(2), new ContextAttribs(3, 3));
					} catch (LWJGLException e) {
						e.printStackTrace();
					}
				}
				if (Keyboard.isKeyDown(Keyboard.KEY_I)) {
					if (!I_down) {
						I_down = true;
						
						System.out.println("Information at " + simulation.time * 1e-9f + "s");
						
						double E = 0, xx, yy, zz;
						
						for (int j = 0; j < simulation.numParticles; j++) {
							xx = simulation.velocity[j * 3 + 0];
							yy = simulation.velocity[j * 3 + 1];
							zz = simulation.velocity[j * 3 + 2];
							E += simulation.mass[j] * (xx * xx + yy * yy + zz * zz) / 2;
						}
						
						System.out.println("Average energy: " + E / simulation.numParticles);
						
						xx = yy = zz = 0;
						
						for (int j = 0; j < simulation.numParticles; j++) {
							xx += simulation.mass[j] * simulation.velocity[j * 3 + 0];
							yy += simulation.mass[j] * simulation.velocity[j * 3 + 1];
							zz += simulation.mass[j] * simulation.velocity[j * 3 + 2];
						}
						
						System.out.println("Average momentum: (" + xx / simulation.numParticles + ", " + yy / simulation.numParticles + ", " + zz / simulation.numParticles + ")");
						
						xx = yy = zz = 0;
						
						for (int j = 0; j < simulation.numParticles; j++) {
							xx += simulation.velocity[j * 3 + 0];
							yy += simulation.velocity[j * 3 + 1];
							zz += simulation.velocity[j * 3 + 2];
						}
						
						System.out.println("Average velocity: (" + xx / simulation.numParticles + ", " + yy / simulation.numParticles + ", " + zz / simulation.numParticles + ")");
						
						System.out.println();
					}
				} else I_down = false;
				
				rotAddMat.setIdentity();
				if (!freeRotation) {
					Matrix4f.translate(position.negate(null), rotAddMat, rotAddMat);
					Matrix4f.rotate(rotationAdd.z, unitZ, rotAddMat, rotAddMat);
					Matrix4f.rotate(rotationAdd.x, unitX, rotAddMat, rotAddMat);
					Matrix4f.rotate(rotationAdd.y, unitY, rotAddMat, rotAddMat);
					Matrix4f.translate(position, rotAddMat, rotAddMat);
				} else {
					Matrix4f.rotate(-rotationAdd.z, unitZ, rotAddMat, rotAddMat);
					Matrix4f.rotate(-rotationAdd.x, unitX, rotAddMat, rotAddMat);
					Matrix4f.rotate(-rotationAdd.y, unitY, rotAddMat, rotAddMat);
				}
				Matrix4f.mul(rotAddMat, viewMat, viewMat);
				
				Matrix4f.invert(viewMat, viewMatInv);
				Matrix4f.translate(position, viewMat, viewMat);
				move4f.set(move.x, move.y, move.z);
				Matrix4f.transform(viewMatInv, move4f, move4f);
				move.set(move4f.x, move4f.y, move4f.z);
				Vector3f.add(position, move, position);
				Matrix4f.translate(position.negate(null), viewMat, viewMat);
				Shader.particle.setUniformMat4("viewMat", viewMat);
				
				prevTime += 1e9 / 60;
				ticks++;
				
				if (ticks % 60 == 0) {
					double sinceLastSecond = (curTime - prevSecond) * 1e-9;
					prevSecond = curTime;
					long ups = simulation.ups;
					simulation.ups = 0;
					if (simulation.playingSimulation) Display.setTitle("Gravity " + Simulation.version + "    " + fps + " fps    |    " + simulation.filePath + "    " + simulation.numParticles + " particles    " + simulation.time * 1e-9f + "s" + (simulation.framesRequested == Integer.MAX_VALUE ? "" : "  (" + (int) (simulation.time * 1e-9f * 6000 / simulation.framesRequested) + "%)"));
					else Display.setTitle("Gravity " + Simulation.version + "    " + fps + " fps    |    " + simulation.filePath + "    " + simulation.numParticles + " particles    " + simulation.time * 1e-9f + "s" + (simulation.framesRequested == Integer.MAX_VALUE ? "" : "  (" + (int) (simulation.time * 1e-9f * 6000 / simulation.framesRequested) + "%)") + "    |    " + simulation.fps + " fps    " + formatLongNicely((long) (ups / sinceLastSecond)) + " ups    " + formatLongNicely(simulation.upf) + " upf");
					simulation.updates = 0;
					fps = 0;
					
					int error = GL11.glGetError();
					if (error != GL11.GL_NO_ERROR) System.out.println("OpenGL error: " + error);
				}
			}
			
			GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);
			render();
			fps++;
			
			Display.update();
			Display.sync(60);
		}
		
		Simulation.simulation.stop();
	}
	
	private String formatLongNicely(long num) {
		StringBuilder str = new StringBuilder(40);
		long mod;
		for (int i = 0; i < 12; i++) {
			mod = num % 1000;
			str.insert(0, mod);
			num /= 1000;
			if (num == 0) break;
			if (mod < 100) str.insert(0, '0');
			if (mod < 10) str.insert(0, '0');
			str.insert(0, ' ');
		}
		return str.toString();
	}
	
	public static Matrix4f projectionMatrix(float fov, float aspect, float near, float far, Matrix4f matr) {
		if (matr == null) matr = new Matrix4f();
		else matr.setIdentity();
		float y_scale = (float) (1 / Math.tan(Math.PI / 180 * fov / 2) * aspect);
		float frustLen = near - far;
		
		matr.m00 = y_scale / aspect;
		matr.m11 = y_scale;
		matr.m22 = (near + far) / frustLen;
		matr.m23 = -1;
		matr.m32 = 2 * near * far / frustLen;
		matr.m33 = 0;
		
		return matr;
	}
	
	private void initGL() {
		try {
			Display.setDisplayMode(new DisplayMode(width, height));
			if (simulation.playingSimulation) Display.setTitle("Gravity " + Simulation.version + "    --- fps    |    " + simulation.filePath + "    " + simulation.numParticles + " particles    " + simulation.time * 1e-9f + "s" + (simulation.framesRequested == Integer.MAX_VALUE ? "" : "  (" + (int) (simulation.time * 1e-9f * 6000 / simulation.framesRequested) + "%)"));
			else Display.setTitle("Gravity " + Simulation.version + "    --- fps    |    " + simulation.filePath + "    " + simulation.numParticles + " particles    " + simulation.time * 1e-9f + "s" + (simulation.framesRequested == Integer.MAX_VALUE ? "" : "  (" + (int) (simulation.time * 1e-9f * 6000 / simulation.framesRequested) + "%)") + "    |    " + simulation.fps + " fps    --- ups    --- upf");
			Display.create(new PixelFormat().withAlphaBits(8).withSamples(2), new ContextAttribs(3, 3));
			Keyboard.create();
//			Mouse.create();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		GL11.glClearColor(0, 0, 0, 0);
		GL11.glEnable(GL11.GL_BLEND);
		GL11.glEnable(GL13.GL_MULTISAMPLE);
		GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_SRC_COLOR);
		GL11.glDepthFunc(GL11.GL_LEQUAL);
	}
	
	public void destroy() {
		GL30.glDeleteVertexArrays(vaoID);
		GL15.glDeleteBuffers(vboID);
		Shader.cleanUp();
		Keyboard.destroy();
//		Mouse.destroy();
		Display.destroy();
	}
}
