package gravity;

import java.nio.FloatBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.function.Consumer;

import org.lwjgl.BufferUtils;
import org.lwjgl.opengl.GL20;
import org.lwjgl.util.vector.Matrix4f;

public class Shader {
	
	private static final ArrayList<Shader> shaders = new ArrayList<Shader>();
	public static Shader particle;
	
	public static void init() {
		cleanUp();
		particle = ShaderBuilder.start("particle").load("vert").load("geom").load("frag").bindAttribs("position", "color", "mass").attatchAll().linkAndFinish();
	}
	
	private static FloatBuffer matrixBuffer = BufferUtils.createFloatBuffer(16);
	
	private HashMap<String, Integer> uniformLocations = new HashMap<String, Integer>();
	private HashMap<String, Integer> attributeLocations = new HashMap<String, Integer>();
	
	public final int shaderID;
	
	Shader(int id) {
		shaderID = id;
		shaders.add(this);
	}
	
	public int getUniformLocation(String name) {
		Integer loc = uniformLocations.get(name);
		if (loc == null) {
			loc = GL20.glGetUniformLocation(shaderID, name);
			uniformLocations.put(name, loc);
		}
		return loc;
	}
	
	public int getAttribLocation(String name) {
		Integer loc = attributeLocations.get(name);
		if (loc == null) {
			loc = GL20.glGetAttribLocation(shaderID, name);
			attributeLocations.put(name, loc);
		}
		return loc;
	}
	
	public void bindAttribute(int loc, String name) {
		GL20.glBindAttribLocation(shaderID, loc, name);
	}
	
	public void setUniform1i(String name, int a) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform1i(loc, a);
	}
	
	public void setUniform1f(String name, float a) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform1f(loc, a);
	}
	
	public void setUniform2f(String name, float a, float b) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform2f(loc, a, b);
	}
	
	public void setUniform3f(String name, float a, float b, float c) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform3f(loc, a, b, c);
	}
	
	public void setUniform4f(String name, float a, float b, float c, float d) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		GL20.glUniform4f(loc, a, b, c, d);
	}
	
	public void setUniformMat4(String name, Matrix4f matrix) {
		int loc = getUniformLocation(name);
		if (loc == -1) return;
		matrix.store(matrixBuffer);
		matrixBuffer.flip();
		GL20.glUniformMatrix4(loc, false, matrixBuffer);
		matrixBuffer.clear(); // maybe unnecessary
	}
	
	public void enable() {
		GL20.glUseProgram(shaderID);
	}
	
	public void delete() {
		GL20.glDeleteProgram(shaderID);
	}
	
	public static void forEachShader(Consumer<Shader> action) {
		shaders.forEach(s -> {
			s.enable();
			action.accept(s);
		});
	}
	
	public static void cleanUp() {
		GL20.glUseProgram(0);
		for (Shader s : shaders)
			s.delete();
		shaders.clear();
	}
	
}
