package gravity;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.attribute.UserDefinedFileAttributeView;
import java.util.HashMap;
import java.util.Random;
import java.util.TreeSet;

import org.lwjgl.util.vector.Vector3f;

public class Simulation {
	private class GravityActionClassForTheStupidSet {
		long time;
		int thing;//acc[thing*3 + 0|1|2 ]
		
		GravityActionClassForTheStupidSet(long l, int i) {
			time = l;
			thing = i;
		}
	}
	
	public static final String version = "1.0";
	public final int numParticles;
	public final double G, precision;
	public final boolean playingSimulation;
	public final long randomSeed;
	public final int framesRequested;
	public final String filePath;
	protected static final Vector3f unitX = new Vector3f(1, 0, 0), unitY = new Vector3f(0, 1, 0), unitZ = new Vector3f(0, 0, 1);
	public static Simulation simulation;
	
	protected boolean running = true;
	
	protected double[] mass;
	protected double[] position;
	protected double[] velocity;
	private double[] accTotal;
	private double[] acc;
	private long[] lastUpdated;
	
	protected long time;
	private long framesTicked;
	protected volatile long ups = 0;
	protected volatile int updates = 0;
	protected long upf = 0;
	protected float fps = 0;
	
	private TreeSet<GravityActionClassForTheStupidSet> updateList = new TreeSet<>((e1, e2) -> {
		return e1.time == e2.time ? e2.thing - e1.thing : e2.time < e1.time ? 1 : -1;
	});
	
	private Random random;
	
	private OutputHandler outputHandler;
	private InputHandler inputHandler;
	private GravityDisplay display;
	
	protected HashMap<String, String> settings = new HashMap<>();
	
	private Simulation(String[] args) {
		simulation = this;
		
		for (String str : args) {
			int index = str.indexOf('=');
			if (index == -1) {
				settings.put(str.toLowerCase(), "true");
			} else {
				settings.put(str.substring(0, index).toLowerCase(), str.substring(index + 1, str.length()));
			}
		}
		
		filePath = getSetting("file", "test.sim");
		File file = new File(filePath);
		playingSimulation = file.exists() && !getSetting("overwrite", false);
		
		if (playingSimulation) {
			try {
				inputHandler = new FileInputHandler(file);
				
				ByteBuffer meta = inputHandler.getMetaData();
				
				inputHandler.startReading();
				
				settings.put("frames", "" + meta.getInt());
				while (meta.remaining() > 0) {
					int len = meta.get() & 0xff;
					String str1 = new String(meta.array(), meta.position(), len);
					meta.position(meta.position() + len);
					len = meta.get() & 0xff;
					String str2 = new String(meta.array(), meta.position(), len);
					meta.position(meta.position() + len);
					settings.put(str1, str2);
				}
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		this.numParticles = getSetting("part", 1000);
		this.precision = Math.pow(10, getSetting("prec", 8));
		this.G = getSetting("g", 1e-2 / Math.pow(Math.cbrt(numParticles), 2));
		randomSeed = getSetting("seed", new Random().nextLong());
		framesRequested = getSetting("frames", Integer.MAX_VALUE);
		
		if (!playingSimulation) {
			try {
				if (file.exists()) file.delete();
				file.createNewFile();
				
				UserDefinedFileAttributeView view = Files.getFileAttributeView(file.toPath(), UserDefinedFileAttributeView.class);
				ByteBuffer temp = ByteBuffer.allocate(8);
				
				temp.putInt(numParticles).flip();
				view.write("Number of particles", Charset.defaultCharset().encode("" + numParticles));
				temp.clear();
				
				temp.putLong(randomSeed).flip();
				view.write("Random seed", Charset.defaultCharset().encode("" + randomSeed));
				temp.clear();
				
				temp.putDouble(precision).flip();
				view.write("Precision", Charset.defaultCharset().encode("" + precision));
				temp.clear();
				
				temp.putDouble(G).flip();
				view.write("G", Charset.defaultCharset().encode("" + G));
				
				outputHandler = new FileOutputHandler(file);
				ByteBuffer buf = ByteBuffer.allocate(100);
				byte[] bts = "part".getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = Integer.toString(numParticles).getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = "seed".getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = Long.toString(randomSeed).getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = "prec".getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = Double.toString(Math.log10(precision)).getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = "g".getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				bts = Double.toString(G).getBytes();
				buf.put((byte) bts.length);
				buf.put(bts);
				
				buf.flip();
				byte[] meta = new byte[buf.limit()];
				buf.get(meta);
				outputHandler.writeMetaData(meta);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
				System.exit(1);
			} catch (IOException e) {
				e.printStackTrace();
				System.exit(1);
			}
		}
		
		mass = new double[numParticles];
		position = new double[numParticles * 3];
		velocity = new double[numParticles * 3];
		accTotal = new double[numParticles * 3];
		acc = new double[numParticles * numParticles * 3];
		lastUpdated = new long[numParticles];
		
		random = new Random(randomSeed);
		
		this.display = !getSetting("nodisplay", false) ? new GravityDisplay(this) : null;
		
		reset();
	}
	
	private String getSetting(String name, String def) {
		String val = settings.get(name);
		return val != null ? val : def;
	}
	
	private boolean getSetting(String name, boolean def) {
		String val = settings.get(name);
		if (val == null) return def;
		try {
			return val != "false";
		} catch (Exception e) {
			return def;
		}
	}
	
	private int getSetting(String name, int def) {
		String val = settings.get(name);
		if (val == null) return def;
		try {
			return Integer.parseInt(val);
		} catch (Exception e) {
			return def;
		}
	}
	
	private long getSetting(String name, long def) {
		String val = settings.get(name);
		if (val == null) return def;
		try {
			return Long.parseLong(val);
		} catch (Exception e) {
			return def;
		}
	}
	
	private double getSetting(String name, double def) {
		String val = settings.get(name);
		if (val == null) return def;
		try {
			return Double.parseDouble(val);
		} catch (Exception e) {
			return def;
		}
	}
	
	protected void reset() {
		updateList.clear();
		for (int i = 0; i < numParticles; i++) {
			position[i * 3 + 0] = random.nextDouble() * 2 - 1;// * (random.nextDouble() * 2 - 1);
			position[i * 3 + 1] = random.nextDouble() * 2 - 1;// * (random.nextDouble() * 2 - 1);
			position[i * 3 + 2] = random.nextDouble() * 2 - 1;// * (random.nextDouble() * 2 - 1) - 2;
			mass[i] = Math.abs((random.nextDouble() * 2 - 1) * (random.nextDouble() * 2 - 1)) * 9.9 + 0.1;
//			if (i == 0) mass[i] = 1000;
			velocity[i * 3 + 0] = velocity[i * 3 + 1] = velocity[i * 3 + 2] = 0;
//			velocity[i * 3 + 0] = position[i * 3 + 1] / 3;
//			velocity[i * 3 + 1] = -position[i * 3 + 0] / 3;
			accTotal[i * 3 + 0] = accTotal[i * 3 + 1] = accTotal[i * 3 + 2] = 0;
			lastUpdated[i] = 0;
		}
		long timeToAdd;
		for (int i = 0; i < numParticles; i++) {
			for (int j = i + 1; j < numParticles; j++) {
				double dx, dy, dz, drSq, a1, a2;
				dx = position[j * 3 + 0] - position[i * 3 + 0];
				dy = position[j * 3 + 1] - position[i * 3 + 1];
				dz = position[j * 3 + 2] - position[i * 3 + 2];
				drSq = dx * dx + dy * dy + dz * dz;
				a1 = G / drSq / Math.sqrt(drSq);
				a2 = a1 * mass[i];
				a1 *= mass[j];
				
				timeToAdd = (long) ((drSq < 0.001 ? 0.001 : drSq > 10 ? 10 + Math.random() : drSq) * precision);
				
				accTotal[i * 3 + 0] += acc[(j * numParticles + i) * 3 + 0] = a1 * dx;
				accTotal[i * 3 + 1] += acc[(j * numParticles + i) * 3 + 1] = a1 * dy;
				accTotal[i * 3 + 2] += acc[(j * numParticles + i) * 3 + 2] = a1 * dz;
				accTotal[j * 3 + 0] += acc[(i * numParticles + j) * 3 + 0] = -a2 * dx;
				accTotal[j * 3 + 1] += acc[(i * numParticles + j) * 3 + 1] = -a2 * dy;
				accTotal[j * 3 + 2] += acc[(i * numParticles + j) * 3 + 2] = -a2 * dz;
				
				updateList.add(new GravityActionClassForTheStupidSet(timeToAdd, i * numParticles + j));
			}
		}
		
		time = 0;
		framesTicked = 0;
	}
	
	// dt is in ns
	private long tick() {
		long nextTime, endTime = framesTicked * 1_000_000_000 / 60, internalUps = 0;
		framesTicked++;
		GravityActionClassForTheStupidSet entry;
		
		while (true) {
			entry = updateList.first();
			nextTime = entry.time;
			if (nextTime > endTime) nextTime = endTime;
			
			int i = entry.thing / numParticles;
			int j = entry.thing % numParticles;
			double dt1 = (nextTime - lastUpdated[i]) * 1e-9;
			double dt2 = (nextTime - lastUpdated[j]) * 1e-9;
			
			position[i * 3 + 0] += velocity[i * 3 + 0] * dt1 + accTotal[i * 3 + 0] * dt1 * dt1 * 0.5f;
			position[i * 3 + 1] += velocity[i * 3 + 1] * dt1 + accTotal[i * 3 + 1] * dt1 * dt1 * 0.5f;
			position[i * 3 + 2] += velocity[i * 3 + 2] * dt1 + accTotal[i * 3 + 2] * dt1 * dt1 * 0.5f;
			position[j * 3 + 0] += velocity[j * 3 + 0] * dt2 + accTotal[j * 3 + 0] * dt2 * dt2 * 0.5f;
			position[j * 3 + 1] += velocity[j * 3 + 1] * dt2 + accTotal[j * 3 + 1] * dt2 * dt2 * 0.5f;
			position[j * 3 + 2] += velocity[j * 3 + 2] * dt2 + accTotal[j * 3 + 2] * dt2 * dt2 * 0.5f;
			velocity[i * 3 + 0] += accTotal[i * 3 + 0] * dt1;
			velocity[i * 3 + 1] += accTotal[i * 3 + 1] * dt1;
			velocity[i * 3 + 2] += accTotal[i * 3 + 2] * dt1;
			velocity[j * 3 + 0] += accTotal[j * 3 + 0] * dt2;
			velocity[j * 3 + 1] += accTotal[j * 3 + 1] * dt2;
			velocity[j * 3 + 2] += accTotal[j * 3 + 2] * dt2;
			
			lastUpdated[i] = lastUpdated[j] = time = nextTime;
			
			if (time == endTime) break;
			
			updateList.remove(entry);
			internalUps++;
			ups++;
			
			double dx, dy, dz, drSq, a1, a2;
			accTotal[i * 3 + 0] -= acc[(j * numParticles + i) * 3 + 0];
			accTotal[i * 3 + 1] -= acc[(j * numParticles + i) * 3 + 1];
			accTotal[i * 3 + 2] -= acc[(j * numParticles + i) * 3 + 2];
			accTotal[j * 3 + 0] -= acc[(i * numParticles + j) * 3 + 0];
			accTotal[j * 3 + 1] -= acc[(i * numParticles + j) * 3 + 1];
			accTotal[j * 3 + 2] -= acc[(i * numParticles + j) * 3 + 2];
			dx = position[j * 3 + 0] - position[i * 3 + 0];
			dy = position[j * 3 + 1] - position[i * 3 + 1];
			dz = position[j * 3 + 2] - position[i * 3 + 2];
			drSq = dx * dx + dy * dy + dz * dz;
			a1 = G / drSq / Math.sqrt(drSq);
			a2 = a1 * mass[i];
			a1 *= mass[j];
			
			entry.time += (long) ((drSq < 0.001 ? 0.001 : drSq > 10 ? 10 + Math.random() : drSq) * precision);
			
			accTotal[i * 3 + 0] += acc[(j * numParticles + i) * 3 + 0] = a1 * dx;
			accTotal[i * 3 + 1] += acc[(j * numParticles + i) * 3 + 1] = a1 * dy;
			accTotal[i * 3 + 2] += acc[(j * numParticles + i) * 3 + 2] = a1 * dz;
			accTotal[j * 3 + 0] += acc[(i * numParticles + j) * 3 + 0] = -a2 * dx;
			accTotal[j * 3 + 1] += acc[(i * numParticles + j) * 3 + 1] = -a2 * dy;
			accTotal[j * 3 + 2] += acc[(i * numParticles + j) * 3 + 2] = -a2 * dz;
			updateList.add(entry);
		}
		
		double delta;
		for (int i = 0; i < numParticles; i++) {
			delta = (time - lastUpdated[i]) * 1e-9;
			position[i * 3 + 0] += velocity[i * 3 + 0] * delta + accTotal[i * 3 + 0] * delta * delta * 0.5f;
			position[i * 3 + 1] += velocity[i * 3 + 1] * delta + accTotal[i * 3 + 1] * delta * delta * 0.5f;
			position[i * 3 + 2] += velocity[i * 3 + 2] * delta + accTotal[i * 3 + 2] * delta * delta * 0.5f;
			velocity[i * 3 + 0] += accTotal[i * 3 + 0] * delta;
			velocity[i * 3 + 1] += accTotal[i * 3 + 1] * delta;
			velocity[i * 3 + 2] += accTotal[i * 3 + 2] * delta;
			lastUpdated[i] = time;
		}
		
		return internalUps;
	}
	
	private void loop() {
		if (playingSimulation) return;
		long startTime;
		
		while (running) {
			if (framesTicked >= framesRequested) {
				running = false;
				break;
			}
			startTime = System.nanoTime();
			upf = tick();
			fps = 1e9f / (System.nanoTime() - startTime);
			updates++;
			
			sendDataToOutput();
		}
	}
	
	private void sendDataToOutput() {
		ByteBuffer buffer = ByteBuffer.allocate(numParticles * 7 * 4 + 4);
		buffer.putInt(buffer.capacity() - 4);
		for (int i = 0; i < numParticles; i++) {
			buffer.putFloat((float) position[i * 3 + 0]);
			buffer.putFloat((float) position[i * 3 + 1]);
			buffer.putFloat((float) position[i * 3 + 2]);
			buffer.putFloat((float) velocity[i * 3 + 0]);
			buffer.putFloat((float) velocity[i * 3 + 1]);
			buffer.putFloat((float) velocity[i * 3 + 2]);
			buffer.putFloat((float) mass[i]);
		}
		buffer.flip();
		outputHandler.pushBuffer(buffer);
	}
	
	public void loadDataFromInput() {
		framesTicked++;
		time = framesTicked * 1_000_000_000 / 60;
		
		ByteBuffer buffer = inputHandler.getBuffer();
		if (buffer == null) return;
		
		for (int i = 0; i < numParticles; i++) {
			position[i * 3 + 0] = buffer.getFloat();
			position[i * 3 + 1] = buffer.getFloat();
			position[i * 3 + 2] = buffer.getFloat();
			velocity[i * 3 + 0] = buffer.getFloat();
			velocity[i * 3 + 1] = buffer.getFloat();
			velocity[i * 3 + 2] = buffer.getFloat();
			mass[i] = buffer.getFloat();
		}
	}
	
	public void stop() {
		running = false;
		destroy();
	}
	
	public void destroy() {
		if (outputHandler != null) outputHandler.destroy();
		if (inputHandler != null) inputHandler.destroy();
		if (display != null) display.destroy();
	}
	
	public static void main(String[] args) {
		new Simulation(args).loop();
	}
}
